@@ -12,11 +12,99 @@ from importlib import import_module
 from pathlib import Path
 from typing import Mapping, MutableMapping, Optional, Sequence, Tuple, Union
 
+import jinja2
 import tomlkit
+from dunamai import (
+    bump_version,
+    check_version,
+    Concern,
+    Pattern,
+    serialize_pep440,
+    serialize_pvp,
+    serialize_semver,
+    Style,
+    Vcs,
+    Version,
+)
 
 _BYPASS_ENV = "POETRY_DYNAMIC_VERSIONING_BYPASS"
 _OVERRIDE_ENV = "POETRY_DYNAMIC_VERSIONING_OVERRIDE"
 
+if sys.version_info >= (3, 8):
+    from typing import TypedDict
+
+    _SubstitutionPattern = TypedDict(
+        "_SubstitutionPattern",
+        {
+            "value": str,
+            "mode": Optional[str],
+        },
+    )
+
+    _SubstitutionFolder = TypedDict(
+        "_SubstitutionFolder",
+        {
+            "path": str,
+            "files": Optional[Sequence[str]],
+            "patterns": Optional[Sequence[Union[str, _SubstitutionPattern]]],
+        },
+    )
+
+    _Substitution = TypedDict(
+        "_Substitution",
+        {
+            "files": Sequence[str],
+            "patterns": Sequence[Union[str, _SubstitutionPattern]],
+            "folders": Sequence[_SubstitutionFolder],
+        },
+    )
+
+    _File = TypedDict(
+        "_File",
+        {
+            "persistent-substitution": Optional[bool],
+            "initial-content": Optional[str],
+            "initial-content-jinja": Optional[str],
+        },
+    )
+
+    _JinjaImport = TypedDict(
+        "_JinjaImport",
+        {
+            "module": str,
+            "item": Optional[str],
+        },
+    )
+
+    _Config = TypedDict(
+        "_Config",
+        {
+            "enable": bool,
+            "vcs": str,
+            "dirty": bool,
+            "pattern": Optional[str],
+            "latest-tag": bool,
+            "substitution": _Substitution,
+            "files": Mapping[str, _File],
+            "style": Optional[str],
+            "metadata": Optional[bool],
+            "format": Optional[str],
+            "format-jinja": Optional[str],
+            "format-jinja-imports": Sequence[_JinjaImport],
+            "bump": bool,
+            "tagged-metadata": bool,
+            "full-commit": bool,
+            "tag-branch": Optional[str],
+            "tag-dir": str,
+            "strict": bool,
+            "fix-shallow-repository": bool,
+        },
+    )
+else:
+
+    class _Config(Mapping):
+        pass
+
 
 class _ProjectState:
     def __init__(
@@ -24,12 +112,14 @@ class _ProjectState:
         path: Path,
         original_version: str,
         version: str,
-        substitutions: MutableMapping[Path, str] = None,
+        substitutions: Optional[MutableMapping[Path, str]] = None,
     ) -> None:
         self.path = path
         self.original_version = original_version
         self.version = version
-        self.substitutions = {} if substitutions is None else substitutions
+        self.substitutions = (
+            {} if substitutions is None else substitutions
+        )  # type: MutableMapping[Path, str]
 
 
 class _State:
@@ -67,7 +157,7 @@ class _FolderConfig:
         self.patterns = patterns
 
     @staticmethod
-    def from_config(config: Mapping, root: Path) -> Sequence["_FolderConfig"]:
+    def from_config(config: _Config, root: Path) -> Sequence["_FolderConfig"]:
         files = config["substitution"]["files"]
         patterns = _SubPattern.from_config(config["substitution"]["patterns"])
 
@@ -75,8 +165,8 @@ class _FolderConfig:
         extra = [
             _FolderConfig(
                 root / x["path"],
-                x.get("files", files),
-                _SubPattern.from_config(x["patterns"]) if "patterns" in x else patterns,
+                x["files"] if x["files"] is not None else files,
+                _SubPattern.from_config(x["patterns"]) if x["patterns"] is not None else patterns,
             )
             for x in config["substitution"]["folders"]
         ]
@@ -132,7 +222,7 @@ def _deep_merge_dicts(base: Mapping, addition: Mapping) -> Mapping:
     return result
 
 
-def _find_higher_file(*names: str, start: Path = None) -> Optional[Path]:
+def _find_higher_file(*names: str, start: Optional[Path] = None) -> Optional[Path]:
     # Note: We need to make sure we get a pathlib object. Many tox poetry
     # helpers will pass us a string and not a pathlib object. See issue #40.
     if start is None:
@@ -146,7 +236,7 @@ def _find_higher_file(*names: str, start: Path = None) -> Optional[Path]:
     return None
 
 
-def _get_pyproject_path(start: Path = None) -> Optional[Path]:
+def _get_pyproject_path(start: Optional[Path] = None) -> Optional[Path]:
     return _find_higher_file("pyproject.toml", start=start)
 
 
@@ -164,15 +254,39 @@ def _get_pyproject_path_from_poetry(pyproject) -> Path:
         raise RuntimeError("Unable to determine pyproject.toml path from Poetry instance")
 
 
-def _get_config(local: Mapping) -> Mapping:
-    return _deep_merge_dicts(_default_config(), local)["tool"]["poetry-dynamic-versioning"]
+def _get_config(local: Mapping) -> _Config:
+    def initialize(data, key):
+        if isinstance(data, dict) and key not in data:
+            data[key] = None
+
+    if isinstance(local, tomlkit.TOMLDocument):
+        local = local.unwrap()
+
+    merged = _deep_merge_dicts(_default_config(), local)["tool"][
+        "poetry-dynamic-versioning"
+    ]  # type: _Config
+
+    # Add default values so we don't have to worry about missing keys
+    for x in merged["files"].values():
+        initialize(x, "initial-content")
+        initialize(x, "initial-content-jinja")
+        initialize(x, "persistent-substitution")
+    for x in merged["format-jinja-imports"]:
+        initialize(x, "item")
+    for x in merged["substitution"]["folders"]:
+        initialize(x, "files")
+        initialize(x, "patterns")
+    for x in merged["substitution"]["patterns"]:
+        initialize(x, "mode")
+
+    return merged
 
 
-def _get_config_from_path(start: Path = None) -> Mapping:
+def _get_config_from_path(start: Optional[Path] = None) -> Mapping:
     pyproject_path = _get_pyproject_path(start)
     if pyproject_path is None:
         return _default_config()["tool"]["poetry-dynamic-versioning"]
-    pyproject = tomlkit.parse(pyproject_path.read_text(encoding="utf-8"))
+    pyproject = tomlkit.parse(pyproject_path.read_bytes().decode("utf-8"))
     result = _get_config(pyproject)
     return result
 
@@ -182,7 +296,7 @@ def _validate_config(config: Optional[Mapping] = None) -> Sequence[str]:
         pyproject_path = _get_pyproject_path()
         if pyproject_path is None:
             raise RuntimeError("Unable to find pyproject.toml")
-        config = tomlkit.parse(pyproject_path.read_text(encoding="utf-8"))
+        config = tomlkit.parse(pyproject_path.read_bytes().decode("utf-8"))
 
     return _validate_config_section(
         config.get("tool", {}).get("poetry-dynamic-versioning", {}),
@@ -221,6 +335,45 @@ def _format_timestamp(value: Optional[dt.datetime]) -> Optional[str]:
     return value.strftime("%Y%m%d%H%M%S")
 
 
+def _render_jinja(
+    version: Version, template: str, config: _Config, extra: Optional[Mapping] = None
+) -> str:
+    if extra is None:
+        extra = {}
+
+    if config["bump"] and version.distance > 0:
+        version = version.bump()
+    default_context = {
+        "base": version.base,
+        "version": version,
+        "stage": version.stage,
+        "revision": version.revision,
+        "distance": version.distance,
+        "commit": version.commit,
+        "dirty": version.dirty,
+        "branch": version.branch,
+        "branch_escaped": _escape_branch(version.branch),
+        "timestamp": _format_timestamp(version.timestamp),
+        "env": os.environ,
+        "bump_version": bump_version,
+        "tagged_metadata": version.tagged_metadata,
+        "serialize_pep440": serialize_pep440,
+        "serialize_pvp": serialize_pvp,
+        "serialize_semver": serialize_semver,
+        **extra,
+    }
+    custom_context = {}  # type: dict
+    for entry in config["format-jinja-imports"]:
+        if "module" in entry:
+            module = import_module(entry["module"])
+            if entry["item"] is not None:
+                custom_context[entry["item"]] = getattr(module, entry["item"])
+            else:
+                custom_context[entry["module"]] = module
+    serialized = jinja2.Template(template).render(**default_context, **custom_context)
+    return serialized
+
+
 def _run_cmd(command: str, codes: Sequence[int] = (0,)) -> Tuple[int, str]:
     result = subprocess.run(
         shlex.split(command),
@@ -258,9 +411,9 @@ def _get_override_version(name: Optional[str], env: Optional[Mapping] = None) ->
     return None
 
 
-def _get_version_from_dunamai(vcs, pattern, config, *, strict: Optional[bool] = None):
-    from dunamai import Version
-
+def _get_version_from_dunamai(
+    vcs: Vcs, pattern: Union[str, Pattern], config: _Config, *, strict: Optional[bool] = None
+) -> Version:
     return Version.from_vcs(
         vcs,
         pattern,
@@ -272,30 +425,17 @@ def _get_version_from_dunamai(vcs, pattern, config, *, strict: Optional[bool] =
     )
 
 
-def _get_version(config: Mapping, name: Optional[str] = None) -> str:
+def _get_version(config: _Config, name: Optional[str] = None) -> Tuple[str, Version]:
     override = _get_override_version(name)
     if override is not None:
-        return override
-
-    import jinja2
-    from dunamai import (
-        bump_version,
-        check_version,
-        Concern,
-        Pattern,
-        serialize_pep440,
-        serialize_pvp,
-        serialize_semver,
-        Style,
-        Vcs,
-    )
+        return (override, Version.parse(override))
 
     vcs = Vcs(config["vcs"])
-    style = config["style"]
-    if style is not None:
-        style = Style(style)
+    style = Style(config["style"]) if config["style"] is not None else None
 
-    pattern = config["pattern"] if config["pattern"] is not None else Pattern.Default
+    pattern = (
+        config["pattern"] if config["pattern"] is not None else Pattern.Default
+    )  # type: Union[str, Pattern]
 
     if config["fix-shallow-repository"]:
         # We start without strict so we can inspect the concerns.
@@ -315,37 +455,7 @@ def _get_version(config: Mapping, name: Optional[str] = None) -> str:
         print("Warning: {}".format(concern.message()), file=sys.stderr)
 
     if config["format-jinja"]:
-        if config["bump"] and version.distance > 0:
-            version = version.bump()
-        default_context = {
-            "base": version.base,
-            "version": version,
-            "stage": version.stage,
-            "revision": version.revision,
-            "distance": version.distance,
-            "commit": version.commit,
-            "dirty": version.dirty,
-            "branch": version.branch,
-            "branch_escaped": _escape_branch(version.branch),
-            "timestamp": _format_timestamp(version.timestamp),
-            "env": os.environ,
-            "bump_version": bump_version,
-            "tagged_metadata": version.tagged_metadata,
-            "serialize_pep440": serialize_pep440,
-            "serialize_pvp": serialize_pvp,
-            "serialize_semver": serialize_semver,
-        }
-        custom_context = {}  # type: dict
-        for entry in config["format-jinja-imports"]:
-            if "module" in entry:
-                module = import_module(entry["module"])
-                if "item" in entry:
-                    custom_context[entry["item"]] = getattr(module, entry["item"])
-                else:
-                    custom_context[entry["module"]] = module
-        serialized = jinja2.Template(config["format-jinja"]).render(
-            **default_context, **custom_context
-        )
+        serialized = _render_jinja(version, config["format-jinja"], config)
         if style is not None:
             check_version(serialized, style)
     else:
@@ -358,7 +468,7 @@ def _get_version(config: Mapping, name: Optional[str] = None) -> str:
             tagged_metadata=config["tagged-metadata"],
         )
 
-    return serialized
+    return (serialized, version)
 
 
 def _substitute_version(name: str, version: str, folders: Sequence[_FolderConfig]) -> None:
@@ -377,7 +487,7 @@ def _substitute_version(name: str, version: str, folders: Sequence[_FolderConfig
                 files[resolved] = folder
 
     for file, config in files.items():
-        original_content = file.read_text(encoding="utf-8")
+        original_content = file.read_bytes().decode("utf-8")
         new_content = _substitute_version_in_text(version, original_content, config.patterns)
         if original_content != new_content:
             _state.projects[name].substitutions[file] = original_content
@@ -415,9 +525,9 @@ def _substitute_version_in_text(version: str, content: str, patterns: Sequence[_
 
 
 def _apply_version(
-    version: str, config: Mapping, pyproject_path: Path, retain: bool = False
+    version: str, instance: Version, config: _Config, pyproject_path: Path, retain: bool = False
 ) -> None:
-    pyproject = tomlkit.parse(pyproject_path.read_text(encoding="utf-8"))
+    pyproject = tomlkit.parse(pyproject_path.read_bytes().decode("utf-8"))
 
     pyproject["tool"]["poetry"]["version"] = version  # type: ignore
 
@@ -433,7 +543,20 @@ def _apply_version(
 
     for file_name, file_info in config["files"].items():
         full_file = pyproject_path.parent.joinpath(file_name)
-        if "initial-content" in file_info:
+
+        if file_info["initial-content-jinja"] is not None:
+            if not full_file.parent.exists():
+                full_file.parent.mkdir()
+            initial = textwrap.dedent(
+                _render_jinja(
+                    instance,
+                    file_info["initial-content-jinja"],
+                    config,
+                    {"formatted_version": version},
+                )
+            )
+            full_file.write_bytes(initial.encode("utf-8"))
+        elif file_info["initial-content"] is not None:
             if not full_file.parent.exists():
                 full_file.parent.mkdir()
             initial = textwrap.dedent(file_info["initial-content"])
@@ -466,7 +589,7 @@ def _get_and_apply_version(
             raise RuntimeError("Unable to find pyproject.toml")
 
     if pyproject is None:
-        pyproject = tomlkit.parse(pyproject_path.read_text(encoding="utf-8"))
+        pyproject = tomlkit.parse(pyproject_path.read_bytes().decode("utf-8"))
 
     if name is None or original is None:
         name = pyproject["tool"]["poetry"]["name"]
@@ -482,7 +605,7 @@ def _get_and_apply_version(
     target_dir = pyproject_path.parent
     os.chdir(str(target_dir))
     try:
-        version = _get_version(config, name)
+        version, instance = _get_version(config, name)
     finally:
         os.chdir(str(initial_dir))
 
@@ -490,27 +613,21 @@ def _get_and_apply_version(
     if name is not None and original is not None:
         _state.projects[name] = _ProjectState(pyproject_path, original, version)
         if io:
-            _apply_version(version, config, pyproject_path, retain)
+            _apply_version(version, instance, config, pyproject_path, retain)
 
     return name
 
 
 def _revert_version(retain: bool = False) -> None:
     for project, state in _state.projects.items():
-        pyproject = tomlkit.parse(state.path.read_text(encoding="utf-8"))
-        pyproject["tool"]["poetry"]["version"] = state.original_version  # type: ignore
-
-        if not retain and not _state.cli_mode:
-            pyproject["tool"]["poetry-dynamic-versioning"]["enable"] = True  # type: ignore
-
-        state.path.write_bytes(tomlkit.dumps(pyproject).encode("utf-8"))
+        pyproject = tomlkit.parse(state.path.read_bytes().decode("utf-8"))
 
         if state.substitutions:
             config = _get_config(pyproject)
 
             persistent = []
             for file, file_info in config["files"].items():
-                if file_info.get("persistent-substitution"):
+                if file_info["persistent-substitution"]:
                     persistent.append(state.path.parent.joinpath(file))
 
             for file, content in state.substitutions.items():
@@ -519,4 +636,14 @@ def _revert_version(retain: bool = False) -> None:
 
                 file.write_bytes(content.encode("utf-8"))
 
+            # Reread pyproject.toml in case the substitutions affected it.
+            pyproject = tomlkit.parse(state.path.read_bytes().decode("utf-8"))
+
+        pyproject["tool"]["poetry"]["version"] = state.original_version  # type: ignore
+
+        if not retain and not _state.cli_mode:
+            pyproject["tool"]["poetry-dynamic-versioning"]["enable"] = True  # type: ignore
+
+        state.path.write_bytes(tomlkit.dumps(pyproject).encode("utf-8"))
+
     _state.projects.clear()
