@@ -5,7 +5,6 @@ __all__ = [
 
 import functools
 import os
-import sys
 
 from cleo.commands.command import Command
 from cleo.events.console_command_event import ConsoleCommandEvent
@@ -25,18 +24,19 @@ else:
 
 
 from poetry_dynamic_versioning import (
+    cli,
     _get_config,
     _get_and_apply_version,
     _get_pyproject_path_from_poetry,
     _state,
     _revert_version,
-    _validate_config,
 )
 
 _COMMAND_ENV = "POETRY_DYNAMIC_VERSIONING_COMMANDS"
+_COMMAND_NO_IO_ENV = "POETRY_DYNAMIC_VERSIONING_COMMANDS_NO_IO"
 
 
-def _patch_dependency_versions() -> None:
+def _patch_dependency_versions(io: bool) -> None:
     """
     The plugin system doesn't seem to expose a way to change dependency
     versions, so we patch `Factory.create_poetry()` to do the work there.
@@ -49,7 +49,7 @@ def _patch_dependency_versions() -> None:
     @functools.wraps(Factory.create_poetry)
     def patched_create_poetry(*args, **kwargs):
         instance = original_create_poetry(*args, **kwargs)
-        _apply_version_via_plugin(instance)
+        _apply_version_via_plugin(instance, io=io)
         return instance
 
     Factory.create_poetry = patched_create_poetry
@@ -61,10 +61,26 @@ def _should_apply(command: str) -> bool:
     if override is not None:
         return command in override.split(",")
     else:
-        return command not in ["run", "shell", "dynamic-versioning"]
+        return command not in ["run", "shell", cli.Command.dv, cli.Command.dv_enable]
 
 
-def _apply_version_via_plugin(poetry: Poetry, retain: bool = False, force: bool = False) -> None:
+def _should_apply_with_io(command: str) -> bool:
+    override = os.environ.get(_COMMAND_NO_IO_ENV)
+    if override is not None:
+        return command not in override.split(",")
+    else:
+        return command not in ["version"]
+
+
+def _apply_version_via_plugin(
+    poetry: Poetry,
+    retain: bool = False,
+    force: bool = False,
+    standalone: bool = False,
+    # fmt: off
+    io: bool = True
+    # fmt: on
+) -> None:
     name = _get_and_apply_version(
         name=poetry.local_config["name"],
         original=poetry.local_config["version"],
@@ -72,6 +88,7 @@ def _apply_version_via_plugin(poetry: Poetry, retain: bool = False, force: bool
         pyproject_path=_get_pyproject_path_from_poetry(poetry.pyproject),
         retain=retain,
         force=force,
+        io=io,
     )
     if name:
         version = _state.projects[name].version
@@ -81,13 +98,13 @@ def _apply_version_via_plugin(poetry: Poetry, retain: bool = False, force: bool
         poetry._package._version = PoetryCoreVersion.parse(version)
         poetry._package._pretty_version = version
 
+        if standalone:
+            cli.report_apply(name)
+
 
 class DynamicVersioningCommand(Command):
-    name = "dynamic-versioning"
-    description = (
-        "Apply the dynamic version to all relevant files and leave the changes in-place."
-        " This allows you to activate the plugin behavior on demand and inspect the result."
-    )
+    name = cli.Command.dv
+    description = cli.Help.main
 
     def __init__(self, application: Application):
         super().__init__()
@@ -95,7 +112,23 @@ class DynamicVersioningCommand(Command):
 
     def handle(self) -> int:
         _state.cli_mode = True
-        _apply_version_via_plugin(self._application.poetry, retain=True, force=True)
+        _apply_version_via_plugin(
+            self._application.poetry, retain=True, force=True, standalone=True
+        )
+        return 0
+
+
+class DynamicVersioningEnableCommand(Command):
+    name = cli.Command.dv_enable
+    description = cli.Help.enable
+
+    def __init__(self, application: Application):
+        super().__init__()
+        self._application = application
+
+    def handle(self) -> int:
+        _state.cli_mode = True
+        cli.enable()
         return 0
 
 
@@ -107,7 +140,10 @@ class DynamicVersioningPlugin(ApplicationPlugin):
         self._application = application
 
         application.command_loader.register_factory(
-            "dynamic-versioning", lambda: DynamicVersioningCommand(application)
+            cli.Command.dv, lambda: DynamicVersioningCommand(application)
+        )
+        application.command_loader.register_factory(
+            cli.Command.dv_enable, lambda: DynamicVersioningEnableCommand(application)
         )
 
         try:
@@ -116,11 +152,7 @@ class DynamicVersioningPlugin(ApplicationPlugin):
             # We're not in a Poetry project directory
             return
 
-        errors = _validate_config(local)
-        if errors:
-            print("poetry-dynamic-versioning configuration issues:", file=sys.stderr)
-            for error in errors:
-                print("  - {}".format(error), file=sys.stderr)
+        cli.validate(standalone=False, config=local)
 
         config = _get_config(local)
         if not config["enable"]:
@@ -137,8 +169,10 @@ class DynamicVersioningPlugin(ApplicationPlugin):
         if not _should_apply(event.command.name):
             return
 
-        _apply_version_via_plugin(self._application.poetry)
-        _patch_dependency_versions()
+        io = _should_apply_with_io(event.command.name)
+
+        _apply_version_via_plugin(self._application.poetry, io=io)
+        _patch_dependency_versions(io)
 
     def _revert_version(
         self, event: ConsoleCommandEvent, kind: str, dispatcher: EventDispatcher
@@ -146,4 +180,7 @@ class DynamicVersioningPlugin(ApplicationPlugin):
         if not _should_apply(event.command.name):
             return
 
+        if not _should_apply_with_io(event.command.name):
+            return
+
         _revert_version()
